import math

from .editor import *

# values can be corrupted into a list of arguments by bash brace expansion
# this function handles converting value that was passed via the cmdline
# if the value is a string not from the commandline, call parse_value directly.
def parse_value_from_bash(value):
    if len(value) == 1:
        return parse_value(value[0])
    try:
        if all(arg[0] == "[" and arg[-1] == "]" for arg in value):
            # the user has passed [value] and bash has messed with it.
            return parse_bash_brackets_value(value)
        else:
            # the user passed {value} and bash has messed with it
            val = ",".join(value)          
            val = f"{{{val}}}"
            return parse_value(val)
    except Exception as err:
        if not isinstance(value,str) and any(v == '=' for v in value):
            raise EditorError("do not use '=' when assigning a key to a value") from err
        raise EditorError("too many arguments were passed or bash has preprocessed and mangled your input. Try putting VALUE in quotes.") from err

# takes in elements from bash and returns the value as a parsed python object
# performs a series of checks and passes work to other functions
def parse_bash_brackets_value(elements):
    if any(x[:x.find(',')].count(':')>1 for x in elements):
        # there are subkeys in this object
        return parse_nested_object_list(elements)
    elif all(x.count(',') == 0 for x in elements):
        # no commas in any element means there is only one element in the list
        return parse_split_object(elements)
    else:
        return parse_brackets_semetric(elements)

def parse_split_object(attrs):
    # the object has been split, so {a:1,b:2} is now
    # [a:1]     [b:2]
    obj = {}
    for a in attrs:
        a = a[1:-1]
        k,v = a.split(":",1)
        obj[k] = parse_value(v)
    return [obj]

# given a bash expanded list, determine how many objects were passed and how many
# elements each object had. assumes all elements were of the same attr length
def get_obj_attr_count(elements):
    num_obj = elements[0].count(",") + 1
    #length = num_attrs^ num_obj
    num_attrs = math.log(len(elements)) / math.log(num_obj)
    return num_obj,int(round(num_attrs))


def parse_brackets_semetric(elements):
    n_obj,n_attr = get_obj_attr_count(elements)
    tree = []
    for e in elements:
        subobjs = e[1:-1].split(',')
        elem = []
        for s in subobjs:
            kvp = s.split(':')
            elem.append(kvp)
        tree.append(elem)
    print(tree)

    fixed_input = []
    for obj_i in range(n_obj):
        obj = {}
        for atr_i in range(n_attr):
            # rem_obj is how many objects 
            tree_index = atr_i*n_attr**((n_obj-1-obj_i))
            k = tree[tree_index][obj_i][0] 
            v = tree[tree_index][obj_i][1]
            obj[k] = parse_value(v)
        fixed_input.append(obj)
    return fixed_input

# parses a list of nested objects, assuming they have the same keys
# it is not possibly (to my knowledge) to parse if the first key at least is not 
# repeated in each object
def parse_nested_object_list(elements):
    fixed_list = []
    elems = []
    for e in elements:
        elem = e
        if e[0] == '[':
            elem = elem[1:]
        if e[-1] == ']' or e[-1] ==',':
            elem = elem[:-1]
        elems.append(elem)
    print(elems)
    ptr = 0
    key = None
    for i,e in enumerate(elems):
        if key == None:
            key = e.split(':',1)[0]
        elif e.split(':',1)[0] == key:
            fixed_list.append(parse_nested_object(elems[ptr:i]))
            ptr = i
    fixed_list.append(parse_nested_object(elems[ptr:]))
    return fixed_list

def parse_nested_object(elements):
    obj = {}
    for elem in elements:
        if ':' in elem:
            k,v = elem.split(':',1)
            if k not in obj:
                obj[k] = parse_value(v)
            else:
                obj[k] = [obj[k],parse_value(v)]
    for k in obj:
        if type(obj[k]) == list:
           obj[k] = parse_nested_object(obj[k])
    return obj



#!/usr/bin/env python3
import json
import sys
import click

from .editor import *
from .bash import *

VERSION = "0.1.0"

class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail('Too many matches: {}'.format(', '.join(sorted(matches))))

@click.group(cls=AliasedGroup)
@click.argument('file',type=click.Path(exists=True))
@click.pass_context
@click.version_option(version=VERSION)
def main(ctx,file):
    ctx.obj = file

@main.command(short_help='add a new element or append to a list')
@click.argument('query')
@click.argument('value', nargs=-1, required=True)
@click.option('-p','--preview', is_flag=True,help="preview change without writing to file")
@click.option('-d','--debug', is_flag=True,help="get debugging information")
@click.pass_obj
def add(file,query,value,preview,debug):
    try:
        value = parse_value_from_bash(value)
        with open(file,'r+') as f:
            json_obj = json.load(f)         
            obj,key = query_object(json_obj,query)
            if key == None:
                # multiple returns
                for o,key in obj:
                    add_func(o,key,value)    
            else:
                add_func(obj,key,value)
            if preview:
                print(json.dumps(json_obj,indent=4))
                return
            save_json(f,json_obj)
    except EditorError as err:
            print_err(err)
            if debug:
                print("Debug enabled. Stack Trace:\n")
                raise err

@main.command(short_help='change an existing key to a new value')
@click.argument('query')
@click.argument('value', nargs=-1, required=True)
@click.option('-p','--preview', is_flag=True,help="preview change without writing to file")
@click.option('-d','--debug', is_flag=True,help="get debugging information")
@click.pass_obj
def edit(file,query,value,preview,debug):
    try:
        value = parse_value_from_bash(value)
        with open(file,'r+') as f:
            json_obj = json.load(f)         
            obj,key = query_object(json_obj,query)
            if key == None:
                # multiple returns
                for o,key in obj:
                    edit_func(o,key,value)    
            else:
                edit_func(obj,key,value)
            if preview:
                print(json.dumps(json_obj,indent=4))
                return
            save_json(f,json_obj)
    except EditorError as err:
        print_err(err)
        if debug:
            print("Debug enabled. Stack Trace:\n")
            raise err

@main.command(short_help='delete an element')
@click.argument('query', nargs=-1)
@click.option('-p','--preview', is_flag=True,help="preview change without writing to file")
@click.option('-d','--debug', is_flag=True,help="get debugging information")
@click.pass_obj
def delete(file,query,preview,debug):
    try:
        with open(file,'r+') as f:
            json_obj = json.load(f)       
            for q in query:
                obj,key = query_object(json_obj,q)
                if key == None:
                    # multiple returns
                    for o,k in obj:
                        delete_func(o,k)    
                else:
                    delete_func(obj,key)
            if preview:
                print(json.dumps(json_obj,indent=4))
                return
            save_json(f,json_obj)
    except EditorError as err:
        print_err(err)
        if debug:
            print("Debug enabled. Stack Trace:\n")
            raise err

def save_json(f, json_obj):
        f.seek(0)
        f.truncate()
        json.dump(json_obj,f,indent=4)

def print_err(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


if __name__ == '__main__':
    main()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jse']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.1,<8.0.0']

entry_points = \
{'console_scripts': ['jse = jse.entry:main']}

setup_kwargs = {
    'name': 'jse',
    'version': '0.1.1',
    'description': 'Quickly edit json files from the command line',
    'long_description': '# jse - JSON Editor [![Build Status](https://travis-ci.org/bjubes/jse.svg?branch=master)](https://travis-ci.org/bjubes/jse) [![Coverage Status](https://coveralls.io/repos/github/bjubes/jse/badge.svg?branch=master)](https://coveralls.io/github/bjubes/jse?branch=master)\n\n\nquickly edit json files from the command line\n\njse is pragmatic and terse. It lets you edit json fast, without needing to care about quotes, types, exact indexes, or any of the stuff that makes json a pain.\n## Usage\n```\n$ jse FILE COMMAND QUERY VALUE\n```\n#### TLDR Version\nedit an existing key: `e` or `edit`\n```\n$ jse config.json edit app.version 0.3.3\n```\nadd a new element: `a` or `add`\n```\n$ jse todo.json add list.shopping {task:eggs,done:false}\n```\ndelete a value: `d` or `delete`\n```\n$ jse problems.json delete problems[99]\n```\nfull [examples with json files](#examples) below\n\n## Installing\n\n```\npip3 install jse\n```\n\n### Running from Source\nRequirements:\n- Python 3.7+\n- [Click](https://pypi.org/project/click/)\n\nSteps:\n1. clone the repository\n2. install click `pip3 install click`\n3. make the run script executable `chmod +x run.py`\n4. place jse on the path `ln -s /path/to/run.py ~/.local/bin/jse`\n\nusing pipenv is recommended if you plan to contribute\n\n## Examples\nAssume we have this json file\n```json\n# example.json\n{\n    "users": [\n        {"name": "alice", "age": 21, "admin": false},\n        {"name": "bob", "age": 57, "admin": true},\n        {"name": "charlie", "age": 37, "admin": false}\n    ]\n}\n```\n\nWe want to delete the user alice using jse. All we need to do is specify `d` or `delete` mode and the path to her `user` object\n```\n$ jse example.json d users[0]\n```\nWe can use both index or dot notation.\n```shell\n$ jse example.json d users.0   #users.first or users.^ also work\n```\n```json\n# example.json\n{\n    "users": [\n        {"name": "bob", "age": 57, "admin": true},\n        {"name": "charlie", "age": 37, "admin": false}\n    ]\n}\n```\nNow lets make charlie an admin. To edit an existing field we use the edit command with `e` or `edit`. Edit takes a key to change and its new value.\n```\n$ jse example.json e users.1.admin true\n```\n```json\n# example.json\n{\n    "users": [\n        {"name": "bob", "age": 57, "admin": true},\n        {"name": "charlie", "age": 37, "admin": true}\n    ]\n}\n```\njse is smart enough to infer datatypes from the command line. it can also accept complex nested objects and arrays in a terse, quote-free format. Lets add a new nested field to the file with `add` or `a`\n```\n$ jse example.json a highscore [{score:32.5,user:bob,metadata:{ip:192.168.1.102,client:firefox}}]\n```\n```json\n{\n    "users": [\n        {"name": "bob", "age": 57, "admin": true},\n        {"name": "charlie", "age": 37, "admin": true}\n    ],\n    "highscore": [\n        {\n            "score": 32.5,\n            "user": "bob",\n            "metadata": {\n                "ip": "192.168.1.102",\n                "client": "firefox"\n            }\n        }\n    ]\n}\n```\njse also understands lists, so we can add new elements to one without needing an explicit index. It will infer we are trying to append from `add` instead of changing the list itself to an object (`edit`)\n```\n$ jse example.json a highscore {score:52,user:charlie}\n```\n```json\n{\n    "users": [\n        {"name": "bob", "age": 57, "admin": true},\n        {"name": "charlie", "age": 37, "admin": true}\n    ],\n    "highscore": [\n        {\n            "score": "32.5",\n            "user": "bob",\n            "metadata": {\n                "ip": "192.168.1.102",\n                "client": "firefox"\n            }\n        },\n        {\n            "score": 52.0,\n            "user": "charlie"\n        }\n    ]\n}\n```\nfirst and last (or `^` and `$`) can also be used as a list index for any operation\n\n```\n$ jse example.json a users.first {name:jon,age:22,admin:false}\n```\n```json\n{\n    "users": [\n        {"name": "jon", "age": 22, "admin":false},\n        {"name": "bob", "age": 57, "admin": true},\n        {"name": "charlie", "age": 37, "admin": true}\n    ],\n    "highscore": [\n        {\n            "score": "32.5",\n            "user": "bob",\n            "metadata": {\n                "ip": "192.168.1.102",\n                "client": "firefox"\n            }\n        },\n        {\n            "score": 52.0,\n            "user": "charlie"\n        }\n    ]\n}\n```\n\njse\'s error messages are informative, because no one wants a generic KeyError\n```\n$ jse example.json a users.0.name "not bob"\n\'name\' already has a value. Use the edit command to modify it\n```\n```\n$ jse example.json d users[2]\nThere is no element with index 2. The largest index is 1\n```\n\nYou can also delete mulitple keys by passing them seperately\n```\n$ jse example.json d users.0.age users.1.age users.2.age\n```\n```json\n{\n    "users": [\n        {\n            "name": "jon",\n            "admin": false,\n        },\n        {\n            "name": "bob",\n            "admin": true,\n        },\n        {\n            "name": "charlie",\n            "admin": true\n        }\n    ]\n    ...\n}\n```\nIf you want to select every element, use the  `*` or `all` operator\n```shell\n$ jse example.json d users.*.age # or users.all.age\n```\n',
    'author': 'Brian Jubelirer',
    'author_email': 'brian2386@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bjubes/jse',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)




import sys
FIRST_EXPR = ['first','^']
LAST_EXPR = ['last','$']
ALL_EXPR = ['all','*']
# use the query to find the sub object we have to modify
# returns an obj and its query OR
# a list of obj-query pairs and None
def query_object(json,query):
    query = query.replace("]","").replace("[", ".").split(".")
    obj = json
    for i,k in enumerate(query[:-1]):
        try:
            obj = obj[k]
        except KeyError as err:
            if k in ALL_EXPR:
                return [query_object(obj,q  + "." +".".join(query[i+1:])) for q in obj.keys()],None
            raise EditorError(f"'{k}' doesn't exist. you can add it with --add") from err
        except TypeError:
            try:
                if k in FIRST_EXPR:
                    k = 0
                elif k in LAST_EXPR:
                    k = -1
                elif k in ALL_EXPR:
                    return [query_object(obj, str(l) + "." +".".join(query[i+1:])) for l in range(len(obj))],None
                obj = obj[int(k)]
            except IndexError as err:
                raise EditorError(f"there is no element with index {k}. The largest index is {len(obj)-1}") from err
    return obj, query[-1]

# convert any value in string representation into a python object or standard type
def parse_value(v):
    if v.isdigit():
        return int(v)
    try:
        return float(v)
    except ValueError:
        pass
    
    v = v.strip()
    if v.lower() == "true":
        return True
    if v.lower()== "false":
        return False
    if v.lower() == "null":
        return None
    
    if v[0] == '[' and v[-1] == ']':
        elems = split_on_root(v[1:-1],',')
        return [parse_value(e) for e in elems]

    if v[0] == '{' and v[-1] == '}':
        if len(v.replace(" ","")) == 2:
            return {}
        obj = {}
        elems = split_on_root(v[1:-1],',')
        for elem in elems:
            key,value = elem.split(":",1)
            obj[key] = parse_value(value)
        return obj
        
    return str(v)

# splits a string using delim, but ignores it inside brackets and braces
def split_on_root(string, delim, openers="[{", closers="}]"):
    elems = []
    ptr = 0
    depth = 0
    for i,c in enumerate(string):
        if c == delim and depth == 0:
            elems.append(string[ptr:i])
            ptr = i+1
        if c in openers:
            depth += 1
        if c in closers:
            depth -= 1
    if ptr != len(string):
        elems.append(string[ptr:])
    return elems

########## operations ############


def add_func(obj,key,value):
    try:
        obj[key]
        #this object already exists, but if its a list lets auto append
        if type(obj[key]) == list:
            obj[key].append(value)
            return
        #its not a list, so we are editing an existing value
        raise EditorError(f"'{key}' already has a value. Use the edit command to modify it")
    except KeyError:
        # the key doesn't exist. this means add is valid
        obj[key] = value
    except TypeError as err:
        # this is a list
        if key.lower() in LAST_EXPR:
            obj.append(value)
            return
        if key.lower() in FIRST_EXPR:
            obj.insert(0,value)
        elif int(key) == len(obj):
            obj.append(value)
        elif int(key) < len(obj):
            raise EditorError(f"'{key}' already has a value. Use the edit command to modify it") from err
        else:
            raise EditorError(f"the list only has {len(obj)} elements. remove the index from your query to add to the end of this list automatically") from err


def edit_func(obj,key,value):
    try:
        if key not in obj:
            if isinstance(obj,list):
                raise TypeError
            elif key.lower() in ALL_EXPR:
                if obj == {}:
                    # the passed object is empty, so the key doesnt exist
                    raise EditorError(f"cannot use '{key}' if no children exist")
                for k in obj.keys():
                    obj[k] = value
                return
            raise EditorError(f"'{key}' doesn't exist. you can add it with --add")
        obj[key] = value
    except TypeError:
        try:
            if key.lower() in FIRST_EXPR:
                key = 0
            elif key.lower() in LAST_EXPR:
               key = -1
            elif key.lower() in ALL_EXPR:
                for i in range(len(obj)):
                    obj[i]= value
                return
            obj[int(key)] = value
        except IndexError as err:
            if len(obj) == 0 and int(key) == 0:
                obj.append(value)
                return
            raise EditorError(f"there is no element with index {key}. The largest index is {len(obj)-1}") from err
    except AttributeError as err:
        #obj.keys() failed bc obj isn't a dict.
        raise EditorError(f"cannot use '{key}' if parent is not an object") from err

def delete_func(obj,key):
    try:
        del obj[key]
    except KeyError as err:
        if key.lower() in ALL_EXPR:
            if obj == {}:
                # the passed object is empty, so the key doesnt exist
                raise EditorError(f"cannot use '{key}' if parent has no children") from err
            for k in list(obj.keys()):
                del obj[k]
            return
        raise EditorError(f"'{key}' doesn't exist.") from err
    except:
        try:
            del obj[int(key)]
        except IndexError as err:
            if len(obj) == 0:
                raise EditorError("the list is already empty") from err
            raise EditorError(f"there is no element with index {key}. The largest index is {len(obj)-1}") from err
        except ValueError as err:
            # this is a list but we gave a non-integer as our key
            if key.lower() in FIRST_EXPR:
                del obj[0]
            elif key.lower() in LAST_EXPR:
                del obj[-1]
            elif key.lower() in ALL_EXPR:
                try:
                    for k in list(obj.keys()):
                        del obj[k]
                except AttributeError as err:
                    #obj.keys() failed bc obj isn't a dict.
                    raise EditorError(f"cannot use '{key}' if parent is not an object") from err
            else:
                raise EditorError(f"'{key}' is not a valid list index.") from err

class EditorError(Exception):
    pass
